#!/usr/bin/env perl

######################################################
# iSPY Project                                       #
# Copyright (c) 2008, Zheng Zhang, Purdue University #
######################################################

use strict;
use vars qw($SCRIPTDIR $TMPDIR $t0 $THRESH $LOG $para $IPLIST $SRCIP_OPT $OUTDIR
            $mapping %probing $AS_issue_cursor $seqRound %unreach $nUnreach %cut $nCut_LBound $nCut_UBound @ASlist %probingIps
            $PIPE_IN1 $PIPE_IN2 $PIPE_OUT1 $PIPE_OUT2 $LOG $PROBELOG $TRLOG $PINGLOG $HITIPS_TR $HITIPS_PING $FAILIPS_TR $FAILIPS_PING $npipe1 $npipe2
            @children_proc $abnormal_term);

$SCRIPTDIR=`dirname $0`; chomp $SCRIPTDIR;

use IO::Select;
use POSIX qw(:sys_wait_h);

my $parse_tr_package="$SCRIPTDIR/parse_probelog.itr.pl";
do "$parse_tr_package" unless
    (! -e "$parse_tr_package" &&
     die "Cannot find parse_probelog.itr.pl in $parse_tr_package\n");

use Time::HiRes qw( gettimeofday tv_interval );

sub usage() {
    print "Usage: ./dProbe.pl (optional:)\n";
    print "             -t tmp_dir -o out_dir  -p threads\n";
    print "             -S source_ip\n";
    print "             (mandatory:) -M mapping_file AS_ips_list\n";
#AS|ips_for_tr|ips_for_ping
}


sub start_daemon()
{
    #Setup outgoing pipes
    system ("rm -f $TMPDIR/fifo1; mkfifo $TMPDIR/fifo1");
    system ("rm -f $TMPDIR/fifo2; mkfifo $TMPDIR/fifo2");

    #Traceroute
    my $pid1 = open ($PIPE_IN1,
          "$SCRIPTDIR/iTraceroute ".
          "-F $TMPDIR/fifo1 -g 1 $SRCIP_OPT -C $para -M $mapping ".
          "2>$TMPDIR/tr.err |");
        #or die "Cannot exec cat fifo1: $!\n";
        #NOTE: this "die" is useless! because open returns before the command 
        #is executed!
    push @children_proc, $pid1;

    #??? don't know if this is true
    #Nice thing about using the above open is that the child traceroute process
    #created as such will be automatically kill as soon as this dProbe.pl is killed,
    #no matter dProbe.pl is in foreground or background.
    print $LOG "[FLOW] PIPE_IN1 done\n";

    #Start ping
    my $pid2 = open ($PIPE_IN2,
          "$SCRIPTDIR/iPing ".
          "-F $TMPDIR/fifo2 -R icmp -g 1 $SRCIP_OPT -C $para ".
          "2>$TMPDIR/ping.err |");
        #or die "Cannot exec cat fifo2: $!\n";
    push @children_proc, $pid2;
    print $LOG "[FLOW] PIPE_IN2 done\n";

    #Open up outgoing pipes
    #NOTE: if no one is openning fifo1, the following open will hang still fifo1 is opened. (reasonable)
    open ($PIPE_OUT1, ">$TMPDIR/fifo1") or die "Cannot open fifo1: $!\n";
    open ($PIPE_OUT2, ">$TMPDIR/fifo2") or die "Cannot open fifo2: $!\n";
    #Make them autoflush
    $| = 1, select $_ for select $PIPE_OUT1;
    $| = 1, select $_ for select $PIPE_OUT2;
    print $LOG "[FLOW] PIPE_OUTs done\n";
}

#return 0 all right. return $i the $i th child exit with nonzero
sub check_children_aliveness ()
{
    my $stiff; my $exitv;
    for my $i (0 .. $#children_proc) {
        $stiff = waitpid($children_proc[$i], &WNOHANG);
        $exitv = $?;
        if ($stiff > 0) {
            die "This can't happen: check_children_aliveness(1)\n" if ($exitv == 0);
            return $i + 1;
        }
        elsif ($stiff < 0) {
            die "This can't happen: check_children_aliveness(2), $stiff\n";
        }
    }
    return 0;
}

sub load_list()
{
    while (my $line = <$IPLIST>) {
        chomp $line;
        
        #parse the ips
        my ($AS, $tr_ips, $ping_ips, $tcp_ips) = split ('\|', $line);
        my @t1 = split (' ', $tr_ips);
        my @t2 = split (' ', $ping_ips);

        #init data structure
        %{$probing{$AS}} = 
            ( 
              tr_ips => [ @t1 ],
              ping_ips => [ @t2 ],
              tr_ips_cursor => -1,
              ping_ips_cursor => -1,
              reach => 0, #0: unreach, 1: reach
              ASpath => [ ],
              recent_reach => 0, #indicating if the "recent_ASpath" is meaningful
              recent_ASpath => [ ], #good AS path when it was reached
              inprobing => 0,
              cut => '',
              );
        push @ASlist, $AS;
    }
    
    $AS_issue_cursor = 0;
    $seqRound = 0;
    
    #Two important sets:
    %unreach = ();
    $nUnreach = 0;
    
    %cut = ();
    $nCut_LBound = 0;
    $nCut_UBound = 0;
}

#Update status of ASes when the AS goes into the pipeline, ie. probing of it starts
sub update_status_ppl_pump($)
{
    my $AS = $_[0];

    die "Cannot find AS $AS in %probing\n" if ! exists $probing{$AS};
    if ($probing{$AS}->{inprobing}) {
        print STDERR "Warning: work on an inprogress AS $AS\n";
    }
    
    #Shift current status into history. Only remeber good path.
    if ($probing{$AS}->{reach}) {
        $probing{$AS}->{recent_reach} = 1;
        my @t = @{$probing{$AS}->{ASpath}};
        $probing{$AS}->{recent_ASpath} = \@t;
        'once_reach' => 0,
    }
    
    $probing{$AS}->{tr_ips_cursor} = -1;
    $probing{$AS}->{ping_ips_cursor} = -1;
    $probing{$AS}->{reach} = 0;
    $probing{$AS}->{ASpath} = [ ];
    $probing{$AS}->{inprobing} = 1;
}

sub pipeline_pump()
{
    #Circular
    die "Bad AS_issue_cursor $AS_issue_cursor\n" 
        if ($AS_issue_cursor < 0 || $AS_issue_cursor > $#ASlist);
    my $AS = $ASlist[$AS_issue_cursor];
    
    update_status_ppl_pump ($AS);

    #debug
    my $now = tv_interval ( $t0, [gettimeofday()]);
    print $LOG "[READ_IN] AS= $AS time= $now\n";
    
    tr_issue($AS);
    
    #Increment cursor
    if ($AS_issue_cursor == $#ASlist) {
        $seqRound ++;
        $AS_issue_cursor = 0;
        print $LOG "[ROUND] round = $seqRound time= $now\n";
    } else {
        $AS_issue_cursor ++;
    }
}

sub tr_issue($)
{
    my $AS = $_[0];

    die "Cannot find AS $AS in %probing\n" if ! exists $probing{$AS};
    my @tr_ips = @{$probing{$AS}{tr_ips}};
    if ($#tr_ips <= $probing{$AS}{tr_ips_cursor}) {
        ping_issue($AS);
        return;
    }

    #Try one more ip
    $probing{$AS}{tr_ips_cursor} ++;
    my $ip = $tr_ips[$probing{$AS}{tr_ips_cursor}];

    $probingIps{$ip} = $AS;

    #really does the issue
    print $PIPE_OUT1 "-A $AS $ip\n";
    $npipe1 ++;
    #debugging
    my $now = tv_interval ( $t0, [gettimeofday()] );
    print $LOG "[TO_TR] npipe1= $npipe1 $ip $AS time= $now\n";
}

sub tr_inspect_result($$)
{
    my ($ip, $text) = @_;
    
    my $AS = $probingIps{$ip};
    die "Cannot find ip $ip in %probingIps\n" if ! defined $AS;
    delete $probingIps{$ip};

    $npipe1 --;
    #debugging
    my $now = tv_interval ( $t0, [gettimeofday()] );
    print $LOG "[FROM_TR] npipe1= $npipe1 $ip $AS time= $now\n";
    
    #Parse the tr path
    my @ippath = (); my @aspath = (); my @reduced_aspath = ();
    my ($ret, $timestamp);
    resolve_ASpath ($text, \@ippath, \@aspath, \$ret, \$timestamp);
    reduce_ASpath (\@aspath, \@reduced_aspath);
    
    print $LOG "[TR_PARSE]: ippath= @ippath\n";
    print $LOG "[TR_PARSE]: aspath= @aspath\n";
    print $LOG "[TR_PARSE]: reduced= @reduced_aspath\n";
    
    die "Cannot find AS $AS in %probing\n" if ! exists $probing{$AS};
    $probing{$AS}->{ASpath} = \@reduced_aspath;
    
    if ($ret eq "enter-AS") {
        print $HITIPS_TR $ip, "\n";
        $probing{$AS}->{reach} = 1;
        pipeline_pop($AS, "tr_good");
    }
    else {
        print $FAILIPS_TR $ip, "\n";
        tr_issue($AS);
    }
}

sub ping_issue($)
{
    my $AS = $_[0];
    
    die "Cannot find AS $AS in %probing\n" if ! exists $probing{$AS};
    my @ping_ips = @{$probing{$AS}{ping_ips}};
    if ($#ping_ips <= $probing{$AS}{ping_ips_cursor}) {
        pipeline_pop($AS, "bad");
        return;
    }

    #Try one more ip
    $probing{$AS}{ping_ips_cursor} ++;
    my $ip = $ping_ips[$probing{$AS}{ping_ips_cursor}];

    $probingIps{$ip} = $AS;

    #really does the issue
    print $PIPE_OUT2  "$ip\n";
    $npipe2 ++;

    #debugging
    my $now = tv_interval ( $t0, [gettimeofday()] );
    print $LOG "[TO_PING] npipe2= $npipe2 $ip $AS time= $now\n";
}

sub ping_inspect_result($$)
{
    my ($ip, $text) = @_;
    
    my $AS = $probingIps{$ip};
    die "Cannot find ip $ip in %probingIps\n" if ! defined $AS;
    delete $probingIps{$ip};
    
    $npipe2 --;
    #debugging
    my $now = tv_interval ( $t0, [gettimeofday()] );
    print $LOG "[FROM_PING] npipe2= $npipe2 $ip $AS time= $now\n";
    
    if ($text =~ /good_probes= (\d+)/ && $1 > 0) {
        print $HITIPS_PING $ip, "\n";
        die "Cannot find AS $AS in %probing\n" if ! exists $probing{$AS};    
        $probing{$AS}->{reach} = 1;
        pipeline_pop($AS, "ping_good");
    }
    else {
        print $FAILIPS_PING $ip, "\n";
        ping_issue($AS);
    }
}

sub define_cut_simple($$) {
    #P=the last non-tailing AS in $current_aspath
    #if P exists in the previous as path, (P, P' next AS) is the cut
    my ($prev_path, $current_path) = @_;
    my ($start, $end, $i);
    for ($i = $#{$current_path}; $i >=0; $i --) {
        if ($current_path->[$i] !~ /\*/) { $start = $current_path->[$i]; last; }
    }
    for my $j (0 .. $#{$prev_path}) {
        if (defined($start) && $prev_path->[$j] eq $start) {
            $end = $prev_path->[$j + 1] if $j < $#${prev_path};
            return "$start $end";
        }
    }

    return "Uncertain";
}

sub delete_unreach($)
{
    my $AS = $_[0];

    if ($probing{$AS}->{recent_reach}) {
        if (exists $unreach{$AS}) {
            delete $unreach{$AS};
            $nUnreach --;
        }
    }
}

sub add_unreach($)
{
    my $AS = $_[0];
    
    if ($probing{$AS}->{recent_reach}) {
        if (! exists $unreach{$AS}) {
            $unreach{$AS} = 1;
            $nUnreach ++;
        }
    }
}

sub delete_cut($)
{
    my $the_cut = $_[0];

    die "not exist the_cut= $the_cut\n" if ! exists $cut{$the_cut};
    $cut{$the_cut} --;
    die "Invalid cut{the_cut}\n" if $cut{$the_cut} < 0;
    if ($cut{$the_cut} == 0) {
        delete $cut{$the_cut};
        $nCut_UBound --;
        if ($the_cut =~ /^\d+ \d+$/) { # certain cut
            $nCut_LBound --;
        }
    }
}

sub add_cut($)
{
    my $the_cut = $_[0];
    
    if (! exists $cut{$the_cut}) {
        $cut{$the_cut} = 1;
        if ($the_cut =~ /^\d+ \d+$/) { # certain cut
            $nCut_LBound ++;
        }
        $nCut_UBound ++;
    }
    else {
        die "Invalid cut{the_cut}\n" if $cut{$the_cut} <= 0;
        $cut{$the_cut} ++;
    }
}

sub update_status_ppl_pop($)
{
    my $AS = $_[0];
    
#    if ($probing{$AS}->{reach} == 0 && $probing{$AS}->{'prev_reach'} != 0) {
#        $nUnreach ++;
#    } elsif ($probing{$AS}->{reach} != 0 && $probing{$AS}->{'prev_reach'} == 0) {
#        $nUnreach --;
#        die "nUnreach < 0!" if ($nUnreach < 0);
#    }
    
    if ($probing{$AS}->{reach}) {
        delete_unreach ($AS);
    }
    else {
        add_unreach ($AS);
    }
    
    #Update cut
    print $LOG "[ANALYZE] AS=$AS old_cut=", $probing{$AS}->{cut}, "\n";
    my ($old_cut, $new_cut);

    #If cause a cut this time
    if (! $probing{$AS}->{reach}) {
        if ($probing{$AS}->{recent_reach}) {
            #Old cut expires, remove
            $old_cut = $probing{$AS}->{cut};
            if ($old_cut ne '') {
                delete_cut ($old_cut);
            }
            
            $new_cut = define_cut_simple ($probing{$AS}->{recent_ASpath}, $probing{$AS}->{ASpath});
            if ($new_cut ne '') {
                add_cut ($new_cut);
            }
            $probing{$AS}->{cut} = $new_cut;

            print $LOG "[ANALYZE] AS=$AS, update or new cut. ASpath=@{$probing{$AS}->{ASpath}}, recent_ASpath=@{$probing{$AS}->{recent_ASpath}}\n";
        }
    }
    else {
        #Old cut expires, remove
        $old_cut = $probing{$AS}->{cut};
        if ($old_cut ne '') {
            delete_cut ($old_cut);
        }
        $probing{$AS}->{cut} = '';
        print $LOG "[ANALYZE] AS=$AS, delete old cut\n";
    }
    
    print $LOG "[ANALYZE] AS=$AS new_cut=", $probing{$AS}->{cut}, "\n";
    
    $probing{$AS}->{inprobing} = 0;
    
    print $LOG "[POP] nUnreach= $nUnreach nCut= [", $nCut_LBound, " ", $nCut_UBound, "]\n";
    
    my $alarm = ($nCut_LBound > $THRESH) ? 1 : 0;
    
    my $now = tv_interval ( $t0, [gettimeofday()]);
    print $PROBELOG "$now $AS status=", ${probing{$AS}{tr_ips_cursor}},
      ",", ${probing{$AS}{ping_ips_cursor}},
      ",", $probing{$AS}->{reach},
      " old_cut=(", $old_cut, ")", " new_cut=(", $new_cut, ")",
      " nUnreach=$nUnreach nCut=[", $nCut_LBound, ",", $nCut_UBound, "]",
      " alarm=", $alarm, "\n";
}

sub pipeline_pop()
{
    my ($AS, $status) = @_;
    
    die "Cannot find AS $AS in %probing\n" if ! exists $probing{$AS};
    
    update_status_ppl_pop ($AS);
    
#was,    delete $probing{$AS};
    $probing{$AS}->{tr_ips_cursor} = -1;
    $probing{$AS}->{ping_ips_cursor} = -1;
    
    #debugging
    my $now = tv_interval ( $t0, [gettimeofday()]);
    print $LOG "[COMMIT] $AS $status time= $now\n";
    pipeline_pump();
}

sub init_pipeline()
{
    $npipe1 = 0;
    $npipe2 = 0;
    for (my $i = 0; $i < $para; $i ++) {
        pipeline_pump();
    }
}

sub handle_out1()
{
    my $committed_ip; my $text = '';
    while (my $out = <$PIPE_IN1>) {
        if ($out =~ /\->([\d\.]+):\d+ /) {
            die "Unpaired START END: START twice in pipe1\n" 
                if defined($committed_ip);
            $committed_ip = $1;
        }
        elsif ($out =~ /^$/) { 
            die "Unpaired START END: END without start in pipe1\n" 
                if ! defined($committed_ip);
            last; 
        } 
        
        $text = $text . $out;
        print $TRLOG $out;
    }
    
    tr_inspect_result ($committed_ip, $text);
}

sub handle_out2()
{
    my $committed_ip; my $text = '';
    while (my $out = <$PIPE_IN2>) {
        if ($out =~ /\->([\d\.]+):\d+ /) {
            die "Unpaired START END: START twice in pipe2\n" 
                if defined($committed_ip);
            $committed_ip = $1;
        }
        elsif ($out =~ /^$/) { 
            die "Unpaired START END: END without start in pipe2\n" 
                if ! defined($committed_ip);
            last; 
        }

        $text = $text . $out;
        print $PINGLOG $out;
    }

    if (defined $committed_ip) {
        ping_inspect_result ($committed_ip, $text);
    }
}

$t0 = [gettimeofday()];

$THRESH = 10;

use Getopt::Std;
my %opts;
getopts('t:o:p:M:P:S:', \%opts);
if ($#ARGV != 0) { usage(); exit (1); }
if (! $opts{M}) { usage(); exit(1); }

$TMPDIR = $opts{t} ? $opts{t} : ".";
$OUTDIR = $opts{o} ? $opts{o} : ".";
$para = $opts{p} ? $opts{p} : 10;
$SRCIP_OPT = $opts{S} ? "-S ". $opts{S} : "";

$mapping = $opts{M};

open ($LOG, "> $OUTDIR/dProbe.log")
    or die "Cannot write file $OUTDIR/dProb.log: $!\n";
#Make them autoflush
$| = 1, select $_ for select $LOG;

open ($PROBELOG, "> $OUTDIR/probe.log")
    or die "Cannot write file $OUTDIR/probel.log: $!\n";
#Make them autoflush
$| = 1, select $_ for select $PROBELOG;

open ($TRLOG, "> $OUTDIR/tr.log")
    or die "Cannot write file $OUTDIR/tr.log: $!\n";
open ($PINGLOG, "> $OUTDIR/ping.log")
    or die "Cannot write file $OUTDIR/ping.log: $!\n";
$| = 1, select $_ for select $TRLOG;
$| = 1, select $_ for select $PINGLOG;

open ($HITIPS_TR, "> $OUTDIR/tr.hitips.log")
    or die "Cannot write file $OUTDIR/tr.hitips.log: $!\n";
open ($FAILIPS_TR, "> $OUTDIR/tr.failips.log")
    or die "Cannot write file $OUTDIR/tr.failips.log: $!\n";
open ($HITIPS_PING, "> $OUTDIR/ping.hitips.log")
    or die "Cannot write file $OUTDIR/ping.hitips.log: $!\n";
open ($FAILIPS_PING, "> $OUTDIR/ping.failips.log")
    or die "Cannot write file $OUTDIR/ping.failips.log: $!\n";

open ($IPLIST, $ARGV[0]) 
    or die "Cannot read file $ARGV[0]: $!\n";

load_list ();
close ($IPLIST);

if ($#ASlist < 0) {
    print "Empty input list. Quit.\n";
    exit (0);
}

#watch dog
$abnormal_term = 0;

my $live;
eval {
    start_daemon();

    my $read_set = new IO::Select();
    $read_set->add($PIPE_IN1);
    $read_set->add($PIPE_IN2);
    
    init_pipeline();

    while($npipe1 != 0 || $npipe2 != 0) {
        #poll status of two children
        if ($live = check_children_aliveness ()) {
            die "Child $live died\n";
        }

        my ($rh_set) = IO::Select->select ($read_set, undef, undef, 2);
        foreach my $rh (@$rh_set) {
            if ($rh == $PIPE_IN1) {
                handle_out1();
            }
            elsif ($rh == $PIPE_IN2) {
                handle_out2();
            }
            else {
                die "Cannot come here\n";
            }
        }
    }
}; 
if($@) { 
    print $@; 
    #kill all chldren
    print "\nKill children processes:\n";
    for my $pid (@children_proc) { system ("children.pl -k -p $pid"); }
    
    if ($live) { 
        $abnormal_term = 1;
    }
}

close ($PIPE_OUT1);
close ($PIPE_OUT2);
close ($LOG);
close ($PROBELOG);
close ($TRLOG);
close ($PINGLOG);
close ($HITIPS_TR);
close ($HITIPS_PING);
close ($FAILIPS_TR);
close ($FAILIPS_PING);

exit ($abnormal_term); #This is necessary to killall traceroute instances??? Not sure
